{-# LANGUAGE FlexibleInstances,
             MultiParamTypeClasses,
             UndecidableInstances,
             GeneralizedNewtypeDeriving,
             FlexibleContexts
  #-}

module Control.Monad.ErrorTP where

import Control.Monad
import Control.Monad.Cont.Class
import Control.Monad.Error.Class
import Control.Monad.Error
import Control.Monad.RWS.Class
import Control.Monad.Reader.Class
import Control.Monad.State.Class
import Control.Monad.Trans
import Control.Monad.Writer.Class
import Control.Monad.MonadStateP
import Control.Monad.MonadErrorP
import EffectCapabilities

{- Observe we don't export ErrorTP, unlike the ErrotT implementation.
   And observe that ErrorTP is not an instance of MonadError
 -}

newtype ErrorTP c e m a = ErrorTP {runETP :: ErrorT e m a} 
        deriving (Functor, Monad, MonadPlus, MonadTrans, MonadCont, MonadIO) 

runErrorTP :: (Error e, Monad m) => ErrorTP c e m a -> m (Either e a)
runErrorTP m = runErrorT (runETP m)

-- Instance of the protected MonadErrorP class
instance (Monad m, Error e) => MonadErrorP c e (ErrorTP (c ()) e m) where
    throwErrorp c l     = checkCapability c `seq` _throw l 
                          where _throw l = ErrorTP $ ErrorT $ return (Left l)
    catchErrorp c m h   = checkCapability c `seq` _catch m h
                          where _catch m h = ErrorTP $ ErrorT $ do
                                       a <- runErrorTP m
                                       case a of
                                            Left  l -> runErrorTP (h l)
                                            Right r -> return (Right r)
    

-- -- ---------------------------------------------------------------------------
-- Instances for other mtl transformers

instance (Error e, MonadRWS r w s m) => MonadRWS r w s (ErrorTP c e m)

instance (Error e, MonadStateP c s m) => MonadStateP c s (ErrorTP (c TCPerm) e m) where
    getp c = lift $ getp c
    putp c = lift . (putp c)

instance (Error e, MonadState s m) => MonadState s (ErrorTP c e m) where
    get = lift get
    put = lift . put

instance (Error e, MonadReader r m) => MonadReader r (ErrorTP c e m) where
    ask       = lift ask
    local f m = ErrorTP $ ErrorT $ local f (runErrorTP m)

instance (Error e, MonadWriter w m) => MonadWriter w (ErrorTP c e m) where
    tell     = lift . tell
    listen m = ErrorTP $ ErrorT $ do
        (a, w) <- listen (runErrorTP m)
        case a of
            Left  l -> return $ Left  l
            Right r -> return $ Right (r, w)
    pass   m = ErrorTP $ ErrorT $ pass $ do
        a <- runErrorTP m
        case a of
            Left  l      -> return (Left  l, id)
            Right (r, f) -> return (Right r, f)
